﻿//using ScannerTest.Common;
using System;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Activation;
using Windows.Devices.Enumeration;
using Windows.Devices.PointOfService;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Navigation;

namespace BarcodeScanningApp
{
    public class ScannerData
    {
        public string id;
        public IScanManager scanManager;
    }

    /// <summary>
    /// State machine for tracking device connects, disconnects, suspends, resumes, and various combinations.
    /// </summary>
    /// <param name="input"></param>
    /// <returns></returns>
    public delegate Task<States> Transition(Inputs input);
    public class State
    {
        public Transition transition;
    }

    public class StateMachine
    {
        /// <summary>
        /// Construct the state machine.
        /// </summary>
        /// <param name="numStates"></param>
        public StateMachine(int numStates)
        {
            this.numStates = numStates;
            int[] lengths = { numStates };
            int[] index = { 0 };

            // Create an array of 'numstates' states
            states = Array.CreateInstance(typeof(State), lengths);
            for (int i = 0; i < numStates; i++)
            {
                index[0] = i;
                states.SetValue(new State(), index);
            }
        }

        /// <summary>
        /// Load a transition in a state
        /// </summary>
        /// <param name="state"></param>
        /// <param name="transition"></param>
        public void SetTransition(States state, Transition transition)
        {
            int[] index = { (int)state };
            State theState = (State)states.GetValue(index);
            theState.transition = transition;
        }

        /// <summary>
        /// Perform the transition
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public Task<States> DoTransition(Inputs input)
        {
            int[] index = { (int)_currentState };
            State stateNow = (State)states.GetValue(index);
            return stateNow.transition(input);
        }

        public States currentState
        {
            set
            {
                _currentState = value;
            }
            get
            {
                return _currentState;
            }
        }
        private int numStates;
        private Array states;
        private States _currentState;
    }

    /// <summary>
    /// Load the 'States' and 'Inputs'enums with values for your states and inputs
    /// </summary>
    public enum States
    {
        disconnected_suspended_inactive,
        disconnected_suspended_active,
        connected_suspended_inactive,
        connected_suspended_active,
        disconnected_resumed_inactive,
        disconnected_resumed_active,
        connected_resumed_inactive,
        connected_resumed_active,
        invalid_last_one,
    }

    public enum Inputs
    {
        disconnect,
        connect,
        suspend,
        resume,
        activate,
        invalid_last_one,
    }

    public sealed partial class App : Application
    {
        public static ScannerData scannerData = new ScannerData();

        /// <summary>
        /// Provides application-specific behavior to supplement the default Application class.
        /// </summary>
        private TransitionCollection transitions;
        public static StateMachine stateMachine;
        private string ScannerIdToRestore = null;
        public static DeviceWatcher watcher;
        private Timer RestoreTimer = null;
        public static bool isSavingScanData;

        public class scannerInfo
        {
            public string ScannerID { get; set; }
            public string ScannerName { get; set; }
            public override string ToString()
            {
                return ScannerName;
            }
        }

        public static scannerInfo[] ScannerInfoArray = new scannerInfo[0];

        /// <summary>
        /// State transition function for each state conforms with 'Transition' delegate.
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public async Task<States> DisconnectedSuspendedInactiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.connect:
                    stateMachine.currentState = States.connected_suspended_inactive;
                    break;
                case Inputs.resume:
                    stateMachine.currentState = States.disconnected_resumed_inactive;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> DisconnectedSuspendedActiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.resume:
                    stateMachine.currentState = States.disconnected_resumed_active;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> DisconnectedResumedInactiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.connect:
                    stateMachine.currentState = States.connected_resumed_inactive;
                    break;
                case Inputs.suspend:
                    stateMachine.currentState = States.disconnected_suspended_inactive;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> DisconnectedResumedActiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.suspend:
                    stateMachine.currentState = States.disconnected_suspended_active;
                    break;
                case Inputs.connect:
                    stateMachine.currentState = States.connected_resumed_active;
                    await scannerData.scanManager.RestoreScanner(ScannerIdToRestore);
                    ScannerIdToRestore = null;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> ConnectedSuspendedInactiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.resume:
                    stateMachine.currentState = States.connected_resumed_inactive;
                    break;
                case Inputs.disconnect:
                    stateMachine.currentState = States.disconnected_suspended_inactive;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> ConnectedSuspendedActiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.resume:
                    stateMachine.currentState = States.connected_resumed_active;
                    await scannerData.scanManager.RestoreScanner(ScannerIdToRestore);
                    ScannerIdToRestore = null;
                    stateMachine.currentState = States.connected_resumed_active;
                    break;
                case Inputs.disconnect:
                    stateMachine.currentState = States.disconnected_suspended_active;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> ConnectedResumedInactiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.suspend:
                    stateMachine.currentState = States.connected_suspended_inactive;
                    break;
                case Inputs.activate:
                    stateMachine.currentState = States.connected_resumed_active;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }

        public async Task<States> ConnectedResumedActiveState(Inputs input)
        {
            switch (input)
            {
                case Inputs.suspend:
                    stateMachine.currentState = States.connected_suspended_active;
                    break;
                case Inputs.disconnect:
                    stateMachine.currentState = States.disconnected_resumed_active;
                    break;
                default:
                    break;
            }

            return stateMachine.currentState;
        }
        // END OF STATE TRANSITION FUNCTIONS

        /// <summary>
        /// Instantiate this application
        /// </summary>
        public App()
        {
            this.InitializeComponent();
            this.Suspending += OnSuspending;
            this.Resuming += OnResuming;
            scannerData.scanManager = new RealScanManager();
            //            scannerData.scanManager.decoderEvent += new DecoderHandler.DecodeEventHandler(Scanner_DataReceived);
            //            scannerData.scanManager.TriggerPressEvent += new TriggerPressHandler.TriggerPressEventHandler(OnDecoderTriggerEventReceived);
            //            scannerData.scanManager.TriggerReleaseEvent += new TriggerReleaseHandler.TriggerReleaseEventHandler(OnDecoderTriggerEventReceived);
            watcher = DeviceInformation.CreateWatcher(BarcodeScanner.GetDeviceSelector());
            watcher.Added += DeviceAdded;
            watcher.Removed += DeviceRemoved;
            watcher.Updated += DeviceUpdated;

            watcher.Start();
            isSavingScanData = false;

            // Instantiate the state machine
            stateMachine = new StateMachine((int)States.invalid_last_one);

            // Load each state's transition function of 'Transition' delegate type.
            stateMachine.SetTransition(States.disconnected_suspended_inactive, new Transition(DisconnectedSuspendedInactiveState));
            stateMachine.SetTransition(States.disconnected_suspended_active, new Transition(DisconnectedSuspendedActiveState));
            stateMachine.SetTransition(States.connected_suspended_inactive, new Transition(ConnectedSuspendedInactiveState));
            stateMachine.SetTransition(States.connected_suspended_active, new Transition(ConnectedSuspendedActiveState));
            stateMachine.SetTransition(States.disconnected_resumed_inactive, new Transition(DisconnectedResumedInactiveState));
            stateMachine.SetTransition(States.disconnected_resumed_active, new Transition(DisconnectedResumedActiveState));
            stateMachine.SetTransition(States.connected_resumed_inactive, new Transition(ConnectedResumedInactiveState));
            stateMachine.SetTransition(States.connected_resumed_active, new Transition(ConnectedResumedActiveState));

            // App always starts with scanners disconnected. DeviceWatcher will notify us when they connect
            stateMachine.currentState = States.disconnected_resumed_inactive;
        }

        /// <summary>
        /// This device watcher function detects and adds a scanner.
        /// </summary>
        /// <param name="devDetector"></param>
        /// <param name="devInfo"></param>
        public void DeviceAdded(DeviceWatcher devDetector, DeviceInformation devInfo)
        {
            string DeviceName = devInfo.Name;

            bool found = false;
            foreach (scannerInfo sia in ScannerInfoArray)
            {
                if (sia.ScannerName == devInfo.Name)
                {
                    found = true;
                    break;
                }
            }
            if (!found)
            {
                Array.Resize(ref ScannerInfoArray, ScannerInfoArray.Length + 1);
                ScannerInfoArray[ScannerInfoArray.Length - 1] = new scannerInfo();

                ScannerInfoArray[ScannerInfoArray.Length - 1].ScannerName = DeviceName;
                ScannerInfoArray[ScannerInfoArray.Length - 1].ScannerID = devInfo.Id;
                stateMachine.DoTransition(Inputs.connect);
                if (scannerData.scanManager.CanScannerBeRestored(devInfo.Id))
                {
                    ScannerIdToRestore = devInfo.Id;
                    stateMachine.DoTransition(Inputs.connect);
                    RestoreTimer = new Timer(RestoreDeviceProc, null, 5000, Timeout.Infinite);
                }
                else if (devInfo.Id.Contains("HID"))
                {
                    stateMachine.DoTransition(Inputs.connect);
                }
                else
                {
                    stateMachine.DoTransition(Inputs.connect);
                }
            }
            else
            {
                stateMachine.DoTransition(Inputs.connect);
            }
        }

        /// <summary>
        /// Timer wakeup to restore device
        /// </summary>
        /// <param name="state"></param>
        private async void RestoreDeviceProc(object state)
        {
            if (ScannerIdToRestore != null)
            {
                await stateMachine.DoTransition(Inputs.activate);
                ScannerIdToRestore = null;
            }

            Timer t = (Timer)state;
            if (t != null)
            {
                t.Dispose();
            }
        }

        /// <summary>
        /// This device watcher function removes a scanner.
        /// </summary>
        /// <param name="devDetector"></param>
        /// <param name="devInfo"></param>
        public void DeviceRemoved(DeviceWatcher devDetector, DeviceInformationUpdate devInfo)
        {
            stateMachine.DoTransition(Inputs.disconnect);
            ScannerIdToRestore = devInfo.Id;
            scannerData.scanManager.ScannerInitializing(false);
            int idxCurrentScanner = Array.FindIndex(ScannerInfoArray, x => x.ScannerID == devInfo.Id);
            if (idxCurrentScanner != -1)
            {
                if (idxCurrentScanner < ScannerInfoArray.Length - 1)
                {
                    Array.Copy(ScannerInfoArray, idxCurrentScanner + 1, ScannerInfoArray, idxCurrentScanner, ScannerInfoArray.Length - idxCurrentScanner - 1);
                }
                Array.Resize(ref ScannerInfoArray, ScannerInfoArray.Length - 1);
                scannerData.scanManager.RemoveScanner(devInfo.Id);
            }
        }

        /// <summary>
        /// Notification that there has been a change in the device
        /// Usually means re-adding it
        /// </summary>
        /// <param name="devDetector"></param>
        /// <param name="devInfo"></param>
        public async void DeviceUpdated(DeviceWatcher devDetector, DeviceInformationUpdate devInfo)
        {
            await stateMachine.DoTransition(Inputs.connect);
        }
        // END DEVICE WATCHER SUPPORT

        /// Invoked when the application is launched normally by the end user.  Other entry points
        /// will be used such as when the application is launched to open a specific file.
        /// </summary>
        /// <param name="e">Details about the launch request and process.</param>
        protected override async void OnLaunched(LaunchActivatedEventArgs e)
        {
#if DEBUG
            if (System.Diagnostics.Debugger.IsAttached)
            {
                this.DebugSettings.EnableFrameRateCounter = true;
            }
#endif
            Frame rootFrame = Window.Current.Content as Frame;

            // Do not repeat app initialization when the Window already has content,
            // just ensure that the window is active
            if (rootFrame == null)
            {
                // Create a Frame to act as the navigation context and navigate to the first page
                rootFrame = new Frame();

                // Associate the frame with a SuspensionManager key.
                SuspensionManager.RegisterFrame(rootFrame, "AppFrame");

                // TODO: Change this value to a cache size that is appropriate for your application.
                rootFrame.CacheSize = 1;

                if (e.PreviousExecutionState == ApplicationExecutionState.Terminated)
                {
                    // Restore the saved session state only when appropriate.
                    try
                    {
                        await SuspensionManager.RestoreAsync();
                    }
                    catch (SuspensionManagerException)
                    {
                        // Something went wrong restoring state.
                        // Assume there is no state and continue.
                    }
                }

                // Place the frame in the current Window
                Window.Current.Content = rootFrame;
            }

            if (rootFrame.Content == null)
            {
                // Removes the turnstile navigation for startup.
                if (rootFrame.ContentTransitions != null)
                {
                    this.transitions = new TransitionCollection();
                    foreach (var c in rootFrame.ContentTransitions)
                    {
                        this.transitions.Add(c);
                    }
                }

                rootFrame.ContentTransitions = null;
                rootFrame.Navigated += this.RootFrame_FirstNavigated;
                
                // When the navigation stack isn't restored navigate to the first page,
                // configuring the new page by passing required information as a navigation
                // parameter.
                if (!rootFrame.Navigate(typeof(MainPage), e.Arguments))
                {
                    throw new Exception("Failed to create initial page");
                }
            }
            // Ensure the current window is active
            Window.Current.Activate();
        }

        /// <summary>
        /// Restores the content transitions after the app has launched.
        /// </summary>
        private void RootFrame_FirstNavigated(object sender, NavigationEventArgs e)
        {
            var rootFrame = sender as Frame;
            rootFrame.ContentTransitions = this.transitions ?? new TransitionCollection() { new NavigationThemeTransition() };
            rootFrame.Navigated -= this.RootFrame_FirstNavigated;
        }

        /// <summary>
        /// Invoked when Navigation to a certain page fails
        /// </summary>
        /// <param name="sender">The Frame which failed navigation</param>
        /// <param name="e">Details about the navigation failure</param>
        void OnNavigationFailed(object sender, NavigationFailedEventArgs e)
        {
            throw new Exception("Failed to load Page " + e.SourcePageType.FullName);
        }

        /// <summary>
        /// This function is called when application execution is being suspended.  Application state is saved
        /// without knowing whether the application will be terminated or resumed with the contents
        /// of memory still intact.
        /// </summary>
        /// <param name="sender">The source of thesuspend request.</param>
        /// <param name="e">Details about the suspend request.</param>
        private async void OnSuspending(object sender, SuspendingEventArgs e)
        {
            var deferral = e.SuspendingOperation.GetDeferral();
            scannerData.scanManager.ScannerInitializing(false);
            await stateMachine.DoTransition(Inputs.suspend);
            ScannerIdToRestore = scannerData.scanManager.GetDeviceId();
            deferral.Complete();
        }

        // This function is called when the app is being resumed.
        private async void OnResuming(object sender, object e)
        {
            // Make device ready for loaded page
            await scannerData.scanManager.RestoreScanner(ScannerIdToRestore);
            await stateMachine.DoTransition(Inputs.resume);
            ScannerIdToRestore = null;
        }
    }
}
