﻿using System;
using Windows.Devices.PointOfService;
using System.Collections.Generic;
using System.Threading.Tasks;
using Windows.Storage.Streams;
using Windows.Media.Playback;

namespace BarcodeScanningApp
{
    /// <summary>
    /// This file contains the code that implements the IScanManager interface.
    /// </summary>
    public sealed class RealScanManager : IScanManager
    {
        private bool triggering = false;
        private BarcodeScanner scanner = null;
        private ClaimedBarcodeScanner claimedBarcodeScanner = null;
        private bool scannerInitializing = false;
        private string scannerID = null;
        private bool bDeviceWasClaimed = false;
        private double savedVolume;
        private IReadOnlyList<uint> savedSymbologyList = null;
        private List<string> activeProfiles = new List<string>();

        public RealScanManager()
        {
            BackgroundMediaPlayer.Current.SetUriSource(new Uri("ms-appx:///assets/Success.wav"));
            savedVolume = BackgroundMediaPlayer.Current.Volume;
            BackgroundMediaPlayer.Current.Volume = 0.0;
            BackgroundMediaPlayer.Current.Play();
        }

        /// <summary>
        /// This function removes a claimed scanner.
        /// </summary>
        /// <param name="_scannerId"></param>
        public void RemoveScanner(string _scannerId)
        {
            try
            {
                // If the scanner to be removed is currently claimed, remove it.
                if ( (_scannerId == scannerID) && (claimedBarcodeScanner != null) )
                {
                    // For HID scanners, don't do a Dispose() on the claimed object after it's been detached. Null out the objects instead.
                    if ( _scannerId.Contains("HID") )
                    {
                        claimedBarcodeScanner = null;
                        scanner = null;
                    }
                    else
                    {
                        claimedBarcodeScanner.Dispose();
                    }
                }
            }
            catch (Exception ex)
            {
                MainPage.Current.NotifyUser("RemoveScanner threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function disables all symbologies for the claimed scanner.
        /// </summary>
        /// <returns></returns>
        public async Task DisableAllSymbologies()
        {
            List<uint> symList = new List<uint>();
            symList.Add(BarcodeSymbologies.Unknown);

            try
            {
                await claimedBarcodeScanner.SetActiveSymbologiesAsync(symList);
                SaveSymbologyList(symList);
            }
            catch (Exception ex)
            {
                throw new Exception("Exception thrown when disabling all symbologies: " + ex.Message);
            }
        }

        /// <summary>
        /// This function restores the scanner (create/claim/enable) after a suspend/resume. It also
        /// restores the symbologies and profiles for the scanner.
        /// </summary>
        /// <param name="_scannerID"></param>
        /// <returns></returns>
        public async Task RestoreScanner(string _scannerID)
        {
            if( !scannerInitializing )
            {
                try
                {
                    scannerInitializing = true;

                    if ((scanner != null) && (scannerID == _scannerID))
                    {
                        // Create the scanner using the saved scanner ID.
                        scanner = null;
                        await CreateScannerByIDObjectAsync(_scannerID);

                        if (bDeviceWasClaimed)
                        {
                            // Claim the scanner.
                            claimedBarcodeScanner = null;
                            if (await ClaimBarcodeScannerAsync())
                            {
                                if (!claimedBarcodeScanner.IsEnabled)
                                {
                                    // Enable the scanner.
                                    if (await EnableBarcodeScannerAsync())
                                    {
                                        // Enable the previously set symbologies and the active profiles.
                                        await claimedBarcodeScanner.SetActiveSymbologiesAsync(GetSavedSymbologyList());
                                        await RestoreActiveProfiles();
                                    }
                                }
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    MainPage.Current.NotifyUser("RestoreScanner threw exception: " + ex.Message);
                }
            }
        }

        /// <summary>
        /// This function returns whether there is a scanner that can be restored.
        /// </summary>
        /// <param name="_scannerID"></param>
        /// <returns></returns>
        public bool CanScannerBeRestored(string _scannerID)
        {
            bool fScannerCanBeRestored = false;

            if( (scanner != null) && (scannerID == _scannerID ) )
            {
                fScannerCanBeRestored = true;
            }

            return fScannerCanBeRestored;
        }

        /// <summary>
        /// This functions sets the scanner intializing state.
        /// </summary>
        /// <param name="initializing"></param>
        public void ScannerInitializing( bool initializing )
        {
            scannerInitializing = initializing;
        }

        /// <summary>
        /// This function returns the scanner initializing state.
        /// </summary>
        /// <returns></returns>
        public bool IsScannerInitialized()
        {
            return scannerInitializing;
        }

        /// <summary>
        /// This function creates the default Barcode Scanner.
        /// </summary>
        /// <returns>a boolean value based on whether it found a compatible scanner connected</returns>
        public async Task<bool> CreateDefaultScannerObjectAsync()
        {
            bool bCreatedStatus = false;

            try
            {
                scanner = await BarcodeScanner.GetDefaultAsync();
                if (null == scanner)
                {
                    MainPage.Current.NotifyUser("Barcode scanner not found. Please connect a barcode scanner.");
                }
                else
                {
                    // Save the scanner ID so it can be restored on a suspend/resume.
                    scannerID = App.scannerData.scanManager.GetDeviceId();
                    bCreatedStatus = true;
                    //                    MainPage.Current.NotifyUser("Default scanner created.");
                }
            }
            catch (Exception ex)
            {
                MainPage.Current.NotifyUser("CreateDefaultScannerObjectAsync threw exception creating scanner: " + ex.Message);
            }

            return bCreatedStatus;
        }

        /// <summary>
        /// This function creates the Barcode Scanner by using the device ID.
        /// </summary>
        /// <returns>a boolean value based on whether it found a compatible scanner connected</returns>
        public async Task<bool> CreateScannerByIDObjectAsync(string deviceId)
        {
            bool bCreatedStatus = false;

            try
            {
                scanner = null;
                scanner = await BarcodeScanner.FromIdAsync(deviceId);
                if (null == scanner)
                {
                    MainPage.Current.NotifyUser("Barcode scanner not found. Please connect a barcode scanner.");
                }
                else
                {
                    // Save the scanner ID so it can be restored on a suspend/resume.
                    scannerID = deviceId;
                    bCreatedStatus = true;
                    //                    MainPage.Current.NotifyUser("Created scanner with " + "Device ID = " + App.scannerData.scanManager.GetDeviceId());
                }
            }
            catch (Exception ex)
            {
                MainPage.Current.NotifyUser("CreateScannerByIDObjectAsync threw exception creating scanner: " + ex.Message);
            }

            return bCreatedStatus;
        }

        /// <summary>
        /// This function claims the created scanner.
        /// </summary>
        /// <returns>a boolean based on whether it was able to claim the scanner.</returns>
        public async Task<bool> ClaimBarcodeScannerAsync()
        {
            bool bClaimAsyncStatus = false;

            if (null != scanner)
            {
                try
                {
                    claimedBarcodeScanner = await scanner.ClaimScannerAsync();

                    if (null == claimedBarcodeScanner)
                    {
                        MainPage.Current.NotifyUser("Claim barcode scanner failed.");
                    }
                    else
                    {
                        // Add the event handlers
                        claimedBarcodeScanner.IsDecodeDataEnabled = true;
                        claimedBarcodeScanner.ReleaseDeviceRequested += claimedBarcodeScanner_ReleaseDeviceRequested;
                        claimedBarcodeScanner.DataReceived += MainPage.Current.Scanner_DataReceived;

                        bClaimAsyncStatus = true;
                        bDeviceWasClaimed = true;
//                        MainPage.Current.NotifyUser("Claimed scanner with " + "Device ID = " + App.scannerData.scanManager.GetDeviceId());
                    }
                }
                catch (Exception ex)
                {
                    MainPage.Current.NotifyUser("Threw exception claiming scanner: " + ex.Message);
                }
            }

            return bClaimAsyncStatus;
        }

        /// <summary>
        /// This function enables the created/claimed scanner.
        /// </summary>
        /// <returns>a boolean based on whether it was able to enable the scanner.</returns>
        public async Task<bool> EnableBarcodeScannerAsync()
        {
            bool bEnableStatus = false;

            if (null != claimedBarcodeScanner)
            {
                try
                {
                    await claimedBarcodeScanner.EnableAsync();

                    if (!claimedBarcodeScanner.IsEnabled)
                    {
                        MainPage.Current.NotifyUser("Enable barcode scanner failed.");
                    }
                    else
                    {
                        bEnableStatus = true;

//                        MainPage.Current.NotifyUser("Enabled scanner with " + "Device ID = " + App.scannerData.scanManager.GetDeviceId());
                    }
                }
                catch (Exception ex)
                {
                    MainPage.Current.NotifyUser("Threw exception enabling scanner: " + ex.Message);
                }
            }

            return bEnableStatus;
        }

        /// <summary>
        /// This function returns the created scanner's device ID.
        /// </summary>
        /// <returns></returns>
        public string GetDeviceId()
        {
            string deviceId = null;

            if (scanner != null)
            {
                deviceId = scanner.DeviceId;
            }
            else
            {
                MainPage.Current.NotifyUser("Failed to get device Id. Scanner instance wasn't created.");
            }

            return deviceId;
        }

        /// <summary>
        /// This function returns the pointer to the created scanner.
        /// </summary>
        /// <returns></returns>
        public BarcodeScanner GetBarcodeScanner()
        {
            return scanner;
        }

        /// <summary>
        /// This function returns the pointer to the claimed scanner.
        /// </summary>
        /// <returns></returns>
        public ClaimedBarcodeScanner GetClaimedBarcodeScanner()
        {
            return claimedBarcodeScanner;
        }

        /// <summary>
        /// This function returns the current created/claimed scanner's ID.
        /// </summary>
        /// <returns></returns>
        public string GetScannerID()
        {
            return scannerID;
        }

        /// <summary>
        /// This function returns the list of enabled symbologies.
        /// </summary>
        /// <returns></returns>
        public IReadOnlyList<uint> GetSavedSymbologyList()
        {
            return savedSymbologyList;
        }

        /// <summary>
        /// This function saves the list of enabled symbologies.
        /// </summary>
        /// <param name="symList"></param>
        public void SaveSymbologyList(IReadOnlyList<uint> symList)
        {
            savedSymbologyList = symList;
        }

        /// <summary>
        /// This function returns the enabled/disabled state of claimed scanner.
        /// </summary>
        /// <returns></returns>
        public bool IsScannerEnabled()
        {
            bool isEnabled = false;

            if( null != claimedBarcodeScanner )
            {
                try
                {
                    isEnabled = claimedBarcodeScanner.IsEnabled;
                }
                catch (Exception ex)
                {
                    throw new Exception("Exception thrown when setting isEnabled. Message=" + ex.Message);
                }
            }
            return isEnabled;
        }

        /// <summary>
        /// This function returns the profiles available with the created scanner.
        /// </summary>
        /// <returns></returns>
        public IReadOnlyList<string> GetSupportedProfiles()
        {
            return scanner.GetSupportedProfiles();
        }

        /// <summary>
        /// This function activates the specified profile and saves it in a list for restoration
        /// after a suspend/resume.
        /// </summary>
        /// <param name="profileName"></param>
        /// <returns></returns>
        public async Task SetActiveProfile(string profileName)
        {
            try
            {
                await claimedBarcodeScanner.SetActiveProfileAsync(profileName);
                activeProfiles.Add(profileName);
            }
            catch (Exception ex)
            {
                throw new Exception("SetActiveProfile threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function restores all active profiles after a suspend/resume.
        /// </summary>
        /// <returns></returns>
        public async Task RestoreActiveProfiles()
        {
            try
            {
                for (int idx = 0; idx < this.activeProfiles.Count; idx++)
                {
                    await claimedBarcodeScanner.SetActiveProfileAsync(activeProfiles[idx]);
                }
            }
            catch (Exception ex)
            {
                throw new Exception("RestoreActiveProfiles threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function clears the list of all previously activated profiles.
        /// </summary>
        public void ClearActiveProfiles()
        {
            activeProfiles.Clear();
        }

        /// <summary>
        /// This function triggers (starts) the barcode scanner.
        /// </summary>
        /// <returns></returns>
        public async Task StartScan()
        {
            try
            {
                if (null != claimedBarcodeScanner)
                {
                    if (claimedBarcodeScanner.IsEnabled)
                    {
                        await claimedBarcodeScanner.StartSoftwareTriggerAsync();
                    }
                }
            }
            catch (Exception ex)
            {
                MainPage.Current.NotifyUser("Threw exception triggering scanner: " + ex.Message);
                throw new Exception("Start scan threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function stops the barcode scanner.
        /// </summary>
        /// <returns></returns>
        public async Task StopScan()
        {
            try
            {
                if (null != claimedBarcodeScanner)
                {
                    if (claimedBarcodeScanner.IsEnabled)
                    {
                        await claimedBarcodeScanner.StopSoftwareTriggerAsync();
                    }
                }
            }
            catch (Exception ex)
            {
                MainPage.Current.NotifyUser("Threw exception untriggering scanner: " + ex.Message);
                throw new Exception("Stop scan threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function resets the scanner by removing the event handlers and disposing the claimed scanner.
        /// </summary>
        public void ResetTheScannerState()
        {
            if (claimedBarcodeScanner != null)
            {
                claimedBarcodeScanner.DataReceived -= MainPage.Current.Scanner_DataReceived;
                claimedBarcodeScanner.ReleaseDeviceRequested -= claimedBarcodeScanner_ReleaseDeviceRequested;
                claimedBarcodeScanner.Dispose();
            }
        }

        /// <summary>
        /// This function disposes the claimed scanner.
        /// </summary>
        public async Task Dispose()
        {
            try
            {
                await StopScan();
                claimedBarcodeScanner.Dispose();
                claimedBarcodeScanner = null;
                bDeviceWasClaimed = false;
            }
            catch (Exception ex)
            {
                claimedBarcodeScanner = null;
                MainPage.Current.NotifyUser("Threw exception disposing scanner: " + ex.Message);
                throw new Exception("Dispose threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// Event handler for releasing the claimed scanner.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void claimedBarcodeScanner_ReleaseDeviceRequested(object sender, ClaimedBarcodeScanner e)
        {
            try
            {
                claimedBarcodeScanner.Dispose();
            }
            catch (Exception ex)
            {
                throw new Exception("Release device requested event threw exception: " + ex.Message);
            }
        }

        /// <summary>
        /// This function gets the current scanner statistics.
        /// </summary>
        /// <param name="statCategories"></param>
        /// <returns></returns>
        public async Task<IBuffer> RetrieveStatistics(string[] statCategories)
        {
            IBuffer statistics = null;
            try
            {
                statistics = await scanner.RetrieveStatisticsAsync(statCategories);
            }
            catch (Exception ex)
            {
                throw new Exception("Retrieve statistics threw exception: " + ex.Message);
            }
            return statistics;
        }

        /// <summary>
        /// This function resets the claimed scanner statistics.
        /// </summary>
        /// <returns></returns>
        public async Task ResetStatistics()
        {
            string[] statCategories = { "" };
            try
            {
                await claimedBarcodeScanner.ResetStatisticsAsync(statCategories);
            }
            catch (Exception ex)
            {
                throw new Exception("Reset statistics threw exception: " + ex.Message);
            }
        }
    }
}