﻿using Honeywell.AIDC.CrossPlatform.Print;
using System;
using System.Threading.Tasks;
using Windows.Storage;
using System.Diagnostics;
using System.Resources;
using System.IO;
using Windows.UI.Core;

namespace LinePrinterShared
{
    // This is shared class which does most of the printing tasks
    class PrintTask
    {
        // Honeywell Line Printer Class
        private LinePrinter linePrinter;
        private static string profiles = null;
        private static string base64ImgStr = null;

        public async Task<bool> DoPrintAsync(IProgress<string> updateUIProgress, 
                                            string printerID, 
                                            string printerURI, 
                                            bool isPrintGraphic)
        {
            bool isSuccessful = true;

            string bytesWritten = "";

            try
            {
                // Read printer command profile into a string
                if (null == profiles)
                {
                    profiles = await readFileFromAssetsFolder(updateUIProgress);
                }

                // Open printer connection
                linePrinter = new LinePrinter(profiles, printerID, printerURI);

                updateUIProgress.Report("- connects printer");

                linePrinter.Connect();

                string sDocNumber = "421701";

                if (isPrintGraphic)
                {
                    updateUIProgress.Report("- writes logo graphic");

                    //Print barista_supply_logo_receipt.bmp logo
                    linePrinter.WriteGraphic("Assets\\barista_supply_logo_receipt.bmp", ImageRotation.DEGREE_0, 0, 0, 0);
                    linePrinter.NewLine(2);
                }

                updateUIProgress.Report("- writes text strings");

                // Set font style to Bold + Double Wide + Double High.
                linePrinter.Bold = true;
                linePrinter.DoubleWide = true;
                linePrinter.DoubleHigh = true;
                linePrinter.Write("SALES ORDER");
                linePrinter.DoubleWide = false;
                linePrinter.DoubleHigh = false;
                linePrinter.NewLine(2);

                // The following text shall be printed in Bold font style.
                linePrinter.Write("CUSTOMER: Casual Step");
                linePrinter.Bold = false;  // Returns to normal font.
                linePrinter.NewLine(2);

                // Set font style to Compressed + Double High.
                linePrinter.DoubleHigh = true;
                linePrinter.Compress = true;
                linePrinter.Write("DOCUMENT#: " + sDocNumber);
                linePrinter.DoubleHigh = false;
                linePrinter.Compress = false;
                linePrinter.NewLine(2);

                // The following text shall be printed in Normal font style.
                linePrinter.Write(" PRD. DESCRIPT.   PRC.  QTY.    NET.");
                linePrinter.NewLine(2);

                linePrinter.Write(" 1501 Timer-Md1  13.15     1   13.15");
                linePrinter.NewLine(1);
                linePrinter.Write(" 1502 Timer-Md2  13.15     3   39.45");
                linePrinter.NewLine(1);
                linePrinter.Write(" 1503 Timer-Md3  13.15     2   26.30");
                linePrinter.NewLine(1);
                linePrinter.Write(" 1504 Timer-Md4  13.15     4   52.60");
                linePrinter.NewLine(1);
                linePrinter.Write(" 1505 Timer-Md5  13.15     5   65.75");
                linePrinter.NewLine(1);
                linePrinter.Write("                        ----  ------");
                linePrinter.NewLine(1);
                linePrinter.Write("              SUBTOTAL    15  197.25");
                linePrinter.NewLine(2);

                linePrinter.Write("          5% State Tax          9.86");
                linePrinter.NewLine(2);

                linePrinter.Write("                              ------");
                linePrinter.NewLine(1);
                linePrinter.Write("           BALANCE DUE        207.11");
                linePrinter.NewLine(1);
                linePrinter.NewLine(1);

                linePrinter.Write(" PAYMENT TYPE: CASH");
                linePrinter.NewLine(2);

                linePrinter.DoubleHigh = true;
                linePrinter.Write("       SIGNATURE / STORE STAMP");
                linePrinter.DoubleHigh = false;
                linePrinter.NewLine(2);

                // Prints the captured signature if it exists.
                if (null == base64ImgStr && isPrintGraphic)
                {
                    base64ImgStr = await GetBase64String(updateUIProgress);
                }

                if(isPrintGraphic && (base64ImgStr.Length > 0))
                {
                    updateUIProgress.Report("- writes graphic with Base64 string");

                    linePrinter.WriteGraphicBase64(base64ImgStr, ImageRotation.DEGREE_0, 0, 350, 100);
                    linePrinter.NewLine(2);
                }

                linePrinter.Write("          ORIGINAL");
                linePrinter.Bold = false;
                linePrinter.NewLine(2);

                updateUIProgress.Report("- writes barcode");

                // Print a Code 39 barcode containing the document number.
                linePrinter.WriteBarcode(BarcodeSymbology.SYMBOLOGY_CODE39,
                        sDocNumber,   // Document# to encode in barcode
                        30,           // Desired height of the barcode in print head dots
                        40);          // Offset in print head dots from the left of the page

                linePrinter.FormFeed();

                bytesWritten = "Number of bytes sent to printer: " + linePrinter.BytesWritten;
            }
            finally
            {
                // release the printer connection
                if (null != linePrinter)
                {
                    linePrinter.Disconnect();
                    linePrinter.Dispose();
                }
            }

            return isSuccessful;
        }

        private async Task<string> readFileFromAssetsFolder(IProgress<string> updateUIProgress)
        {
            updateUIProgress.Report("- read printer_profiles.json file");

            string fileContent = null;

            var file = await StorageFile.GetFileFromApplicationUriAsync(new Uri("ms-appx:///Assets/printer_profiles.json"));
            if (file != null)
            {
                fileContent = await FileIO.ReadTextAsync(file);
                if (fileContent == null)
                    updateUIProgress.Report("- Could not read profile content");
            }
            else
            {
                updateUIProgress.Report("- Could not access printer_profiles.json file");
            }

            return fileContent;
        }

        private async Task<string> GetBase64String(IProgress<string> updateUIProgress)
        {
            updateUIProgress.Report("- read Signature.bmp file.");

            string strBase64Img = string.Empty;
            var imgFile = await StorageFile.GetFileFromApplicationUriAsync
                                            (new Uri("ms-appx:///Assets/Signature.bmp", UriKind.Absolute));
            if (null != imgFile)
            {
                using (BinaryReader fileReader = new BinaryReader(await imgFile.OpenStreamForReadAsync()))
                {
                    var buffer = new byte[fileReader.BaseStream.Length];
                    fileReader.Read(buffer, 0, buffer.Length);
                    strBase64Img = Convert.ToBase64String(buffer);
                }
            }
            else
            {
                updateUIProgress.Report("- Failed on read Signature.bmp file.");
            }

            return strBase64Img;
        }
    }

}
